//---------------------------------------------------------------------------
// Copyright 2005 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING' in this distribution.
//
// $Source: $
// $Id: $
//
// Description: 
//
//---------------------------------------------------------------------------

#ifndef AH_IO_SMART_BUFFER2D_H		// prevent multiple includes
#define AH_IO_SMART_BUFFER2D_H

// --- forward declarations ---

// --- necessary includes ---

#include <AH/Config/Platform.h>
#include <AH/Error/Sys.h>
#include <AH/Error/Intern.h>
USING_NAMESPACE(AH)

#include <cerrno>
#include <string>
#ifdef BOUNDS_CHECK
#   include<iostream>
#endif
USING_NAMESPACE(std)

// --- types and classes ---

BEGIN_NAMESPACE(AH)

/// smart 2D buffer for any type <tt>T</tt>
/**
   Use <tt>AH::IOSmartBuffer</tt> if you need a resizable buffer or
   a buffer which gets properly removed if the scope is left, e.g.
   due to an exception.

   Please include <em> AH/IO/Smart/Buffer.h</em>.
 */
template <class T>
class IOSmartBuffer2D
{
 public:
    /// constructor for buffer of <tt>s</tt> elements of type <tt>T</tt>
    IOSmartBuffer2D(size_t n, size_t s)
	: bnum(n), size(s), max(size), buf (new T*[bnum])
    {
	if (!buf) 
	{
	    bnum = max = size = 0;
	    throw ErrorSys(ENOMEM);
	}
	for (size_t b = 0; b < bnum; ++b)
	{
	    buf[b] = new T[max];
	    if (!buf[b])
	    {
		bnum = max = size = 0;
		throw ErrorSys(ENOMEM);
	    }
	}
    }

    /// constructor for empty buffer of type <tt>T</tt>
    IOSmartBuffer2D(size_t n) : bnum(n), size(0), max(0), buf (new T*[bnum])
    {
	if (!buf) 
	{
	    bnum = 0;
	    throw ErrorSys(ENOMEM);
	}
	for (size_t b = 0; b < bnum; ++b)
	    buf[b] = 0;
    }

    /// destructor
    ~IOSmartBuffer2D()
    {
	if (buf)
	{
	    for (size_t b = bnum; b > 0; --b)
		if (buf[b-1]) delete[] buf[b-1];
	    delete[] buf; 
	}
    }

    /// resize all buffers for <tt>s</tt> elements of type <tt>T</tt>
    /**
      <tt>resize(s)</tt> resizes all buffers to contains <tt>s</tt> elements of 
      type <tt>T</tt>:
      - if <tt>s</tt> is bigger than the current size then the old buffer
        is destroyed and a new one with the appropriate size is created.
      - if <tt>s</tt> is smaller than the current size then nothing is done
        at all.
      - if <tt>s</tt> is equal to the current size then nothing is done
        at all.
      .
      This means that the buffer does only grow but never shrinks. This
      speeds up the execution but blocks memory.
     */
    void resize(size_t s)
    {
	if (s == size) return;
	if (s > max) recreate(s);
	else	     size = s;
    }

    /// resize buffer for <tt>s</tt> elements of type <tt>T</tt>
    /**
      <tt>recreate(s)</tt> resizes the buffer to contains <tt>s</tt> elements of 
      type <tt>T</tt>:
      - if <tt>s</tt> is not equal to the current size then the old buffer
        is destroyed and a new one with the appropriate size is created.
      - if <tt>s</tt> is equal to the current size then nothing is done
        at all.
      .
      This means that the buffer grows and shrinks according to the new 
      size <tt>s</tt>. This costs execution time but does not block memory.
     */
    void recreate(size_t s)
    {
	if (s == size) return;
	if (buf)
	{
	    for (size_t b = bnum; b > 0; --b)
	    {
		if (buf[b-1]) delete[] buf[b-1];
		buf[b] = 0;
	    } 
	}
	else buf = new T*[bnum];
	if (!buf) 
	{
	    bnum = max = size = 0;
	    throw ErrorSys(ENOMEM);
	}
	max = size = s;
	for (size_t b = 0; b < bnum; ++b)
	{
	    buf[b] = new T[max];
	    if (!buf[b])
	    {
		bnum = max = size = 0;
		throw ErrorSys(ENOMEM);
	    }
	}
    }

    /// remove buffer and free its allocated memory
    void remove()
    {
	if (buf)
	{
	    for (size_t b = bnum; b > 0; --b)
		if (buf[b-1]) delete[] buf[b-1];
	    delete[] buf; 
	}
	buf = 0;
	bnum = max = size = 0;
    }

    /// clear all buffer entries (set to 0)
    void clear()
    {
	for (size_t b = 0; b < bnum; ++b)
	{
	    memset(buf[b], 0, getSize());
	}
    }

    /// set all buffer entries to specified value
    void initAll(const T& t)
    {
	for (unsigned int b = 0; b < bnum; ++b)
	    for (unsigned int i = 0; i < size; ++i)
		buf[b][i] = t;
    }

    /// return number of types <tt>T</tt> which can be stored in buffer
    size_t getNum() const { return size; }

    /// return size of buffer in bytes
    size_t getSize() const { return size * sizeof(T); }

    /// return 1D buffer
    T* getBuffer(size_t b)
    {
#ifdef BOUNDS_CHECK
 	if (b < bnum) return buf[b];
	throw ErrorInternBadArg("in IOSmartBuffer::getBuffer()");
#else
 	return buf[b];
#endif
    }

    /// return 1D buffer
    T** getBuffer() { return buf; }

    /// return pointer to buffer
    T** operator()() { return buf; }

    /// return specified entry in buffer (with bound check)
    const T& getEntry(size_t b, size_t idx) const 
    {
#ifdef BOUNDS_CHECK
	if ((b < bnum) && (idx < size)) return buf[b][idx];
	throw ErrorInternBadArg("in IOSmartBuffer::getEntry()");
#else
	return buf[b][idx];
#endif
    }

    /// set specified entry in buffer (with bound check)
    void setEntry(size_t b, size_t idx, const T& t)
    {
#ifdef BOUNDS_CHECK
	if ((b < bnum) && (idx < size)) buf[b][idx] = t;
	else
	{
	    cerr << "b=" << b << "[0.." << bnum-1 << "],"
		 << "i=" << idx << "[0.." << size-1 << "]" << endl;
	    throw ErrorInternBadArg("in IOSmartBuffer::setEntry()");
	}
#else
	buf[b][idx] = t;
#endif
    }

#if 0
    /// return specified readonly entry in buffer (no bound check)
    const T& operator[](size_t idx) const
	{ return buf[idx]; }

    /// return specified writable entry in buffer (no bound check)
    T& operator[](size_t idx)
	{ return buf[idx]; }
#endif
 private:
    size_t	bnum;	// number of buffers
    size_t	size;	// current size of each buffer in number of entries
    size_t	max;	// maximal size of each buffer in number of entries
    T**		buf;	// the buffer itsself

};	// IOSmartBuffer2D


END_NAMESPACE
#endif // !AH_IO_SMART_BUFFER2D_H
