//---------------------------------------------------------------------------
// Copyright 2004 Andreas Huennebeck
//
// Licensed under the terms of the GNU Library General Public License.
// ABSOLUTELY NO WARRANTY. See the file `COPYING.LIB' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/CmdOptHandler.c,v $
// $Id: 
//
// Description: implementation of class IOCmdOption
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: CmdOptHandler.c,v 1.4 2004/05/05 21:16:41 ah Exp $ (c) 2004 Andreas Huennebeck";

// --- necessary includes ---
#include <AH/Config/Platform.h>
#include <AH/IO/CmdOptHandler.h>
#include <AH/IO/CmdOption.h>
#include <AH/Error/Sys.h>
USING_NAMESPACE(AH)

#include <cerrno>
USING_NAMESPACE(std)

// --- local stuff ---

#define TRACE 0

// --- class member implementation ---

BEGIN_NAMESPACE(AH)


void IOCmdOptHandler::addOptions(const std::vector<IOCmdOption*> newOpts)
{
    for (unsigned int i = 0; i < newOpts.size(); ++i)
	opts.push_back(newOpts[i]);
}


// return help text of all options.
std::string IOCmdOptHandler::getHelpText(const std::string& lineHead)
{
    std::string helptext;
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	helptext += lineHead;
	if (opts[i]->isShort())
	{
	    helptext += "-";
	    helptext += opts[i]->getShortOpt();
	}
	if (opts[i]->isLong())
	{
	    if (opts[i]->isShort()) helptext += ", ";
	    helptext += "--";
	    helptext += opts[i]->getLongOpt();
	}
	helptext += "\t";
	helptext += opts[i]->getDescription();
	helptext += "\n";
    }
    return helptext;
}

// return short option string, for help text.
std::string IOCmdOptHandler::getShortOptionText() const
{
    if (0 == short_opt.size()) return "";
    return "[-" + short_opt + "]"; 
}



int IOCmdOptHandler::handleArgs
(
    int argc,
    char* argv[],
    std::vector<std::string>& remarg,
    std::vector<std::string>& errwarn
)
{
    int badArg = 0;
    int a = 1;
    for (; a < argc; ++a)
    {
	if (!argv[a]) 
	{
	    ++badArg;
	    continue;
	}

	if (0 == strncmp(argv[a], "--", 2))
	{
	    // assume long option
	    if (strlen(argv[a]) == 2)
	    {
		errwarn.push_back("incomplete long option '--'");
		++badArg;
		continue;
	    }
	    // it is a long option: check for additional argument:
	    badArg += handleLongArg(a, argc, argv, errwarn);
	}
	else if (argv[a][0] == '-')
	{
	    // assume short option
	    if (strlen(argv[a]) == 1)
		break; // this is stdin or stdout, not an arg
	    // it is a short option
	    badArg += handleShortArg(a, argc, argv, errwarn);
	}
	else break;
    }

    // any remaining stuff are arguments:
    while (a < argc)
	remarg.push_back(argv[a++]);

    return badArg;

}



// return next entry if this is an argument:
std::string IOCmdOptHandler::getNextOptArg(int& a, int argc, char* argv[])
{
    if (a+1 >= argc) return ""; // no more arguments available

    const char* next = argv[a+1];
    if (!next) return ""; // no argument
    if (0 == strcmp(next, "-"))
    {
	++a;
	return "-"; // argument is stdin/out
    }
    if (next[0] == '-') return ""; // this is an option, not an argument

    ++a;
    return next; // this is an argument
}



int IOCmdOptHandler::handleLongArg(int& a, int argc, char* argv[],
				   std::vector<std::string>& errwarn)
{

    int errors = 0;
    bool found = false;

    std::string option(argv[a] + 2);

    // try to find argument contained in long option
    // (e.g "--verbose=4")
    string::size_type pos = option.find_first_of("=",0);
    std::string optarg;
    if (string::npos != pos)
    {
	optarg = option.substr(pos+1);
	option = option.substr(0, pos);
    }

    // loop through all known options:
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	if (opts[i]->getLongOpt() == option) 
	{
	    // option has been found
	    found = true;
	    opts[i]->set();
	    switch (opts[i]->getSecArgMode())
	    {
	    default:
	    case IOCmdOption::SecondArgNone:
		if (optarg != "")
		{
		    errwarn.push_back("option --" + option +
				      " takes no argument");
		    ++errors;
		}
		break;

	    case IOCmdOption::SecondArgAlways:
		if (optarg == "")
		    // try to find argument in next argv 
		    // (e.g "--verbose 4")
		    optarg = getNextOptArg(a, argc, argv);
		if (optarg == "")
		{
		    errwarn.push_back("missing argument for option --" +
				      option);
		    ++errors;
		}
		else opts[i]->setSecArg(optarg);
		break;

	    case IOCmdOption::SecondArgOptional:
		if (optarg == "")
		    optarg = getNextOptArg(a, argc, argv);
		if (optarg != "") opts[i]->setSecArg(optarg);
		break;
	    }
	    break;
	}
    }
    if (!found)
    {
	errwarn.push_back("invalid option --" + option);
	++errors;
    }
    return errors;
}


int IOCmdOptHandler::handleShortArg(int& a, int argc, char* argv[],
	std::vector<std::string>& errwarn)
{
    int errors = 0;
    bool found = false;

    std::string option(argv[a] + 1);

    // try to find argument contained in short option
    // (e.g "-v4")
    std::string optarg;
    if (option.size() > 1)
    {
	optarg = option.substr(1);
	option = option.substr(0, 1);
    }

    // loop through all known options:
    for (unsigned int i = 0; i < opts.size(); ++i)
    {
	if (opts[i]->getShortOpt() == option[0]) 
	{
	    // option has been found
	    found = true;
	    opts[i]->set();
	    switch (opts[i]->getSecArgMode())
	    {
	    default:
	    case IOCmdOption::SecondArgNone:
		if (optarg != "")
		{
		    errwarn.push_back("option -" + option +
				      " takes no argument");
		    ++errors;
		}
		break;

	    case IOCmdOption::SecondArgAlways:
		if (optarg == "")
		    // try to find argument in next argv 
		    // (e.g "-v 4")
		    optarg = getNextOptArg(a, argc, argv);
		if (optarg == "")
		{
		    errwarn.push_back("missing argument for option -" +
				      option);
		    ++errors;
		}
		else opts[i]->setSecArg(optarg);
		break;

	    case IOCmdOption::SecondArgOptional:
		if (optarg == "")
		    optarg = getNextOptArg(a, argc, argv);
		if (optarg != "") opts[i]->setSecArg(optarg);
		break;
	    }
	    break;
	}
    }
    if (!found)
    {
	errwarn.push_back("invalid option -" + option);
	++errors;
    }
    return errors;
}


END_NAMESPACE


