//---------------------------------------------------------------------------
// Copyright 1999 Andreas Huennebeck
//
// Licensed under the terms of the GNU General Public License.
// ABSOLUTELY NO WARRANTY. See the file 'COPYING' in this distribution.
//
// $Source: /home/ah/CvsTree/daten/src/libAHio/RiffWaveFileCached.c,v $
// $Id: 
//
// Description: 
//
//---------------------------------------------------------------------------

static const char rcsid[] = "$Id: RiffWaveFileCached.c,v 1.5 2004/04/20 22:26:22 ah Exp $ (c) 1999 Andreas Huennebeck";

// --- necessary includes ---

#include <AH/IO/Riff/WaveFileCached.h>
#include <AH/Error/Intern.h>
#include <AH/Error/Sys.h>
#include <AH/Error/Basic.h>
USING_NAMESPACE(AH)

#include <cstdio>
#include <string>
USING_NAMESPACE(std)

BEGIN_NAMESPACE(AH)


// constructor for input WAV file 
IORiffWaveFileCached::IORiffWaveFileCached
(
    const char* name,
    uint32 samSize,
    bool isPipe,
    int verbose
)
    : IORiffWaveFile(name, isPipe, verbose),
      cache(0), cacheSamNum(samSize), cacheStart(0), cacheEnd(0), 
      isOutput(false)
{
}



// constructor for input file (WAV, MP3, Ogg, ...) 
IORiffWaveFileCached::IORiffWaveFileCached
(
    const char* name,
    uint32 samSize,
    int verbose
)
    : IORiffWaveFile(name, verbose),
      cache(0), cacheSamNum(samSize), cacheStart(0), cacheEnd(0), 
      isOutput(false)
{
}



// constructor for output WAV file 
IORiffWaveFileCached::IORiffWaveFileCached
(
    const char* name,	// name of output file
    uint16 channel_num,	// number of channels
    uint16 sample_size,	// size of sample in bits
    uint32 sample_rate,	// sample rate in Hz
    uint32 sample_num,  // number of samples per channel
    uint32 samSize,     // cache size in number of samples
    int verbose
)
    : IORiffWaveFile(name, channel_num, sample_size, sample_rate, sample_num,
		     verbose),
      cache(0), cacheSamNum(samSize), cacheStart(0), cacheEnd(0),
      isOutput(true)
{
}



// destructor
IORiffWaveFileCached::~IORiffWaveFileCached()
{
    if (isOutput && cache) writeCache();
    delete[] cache;
}



// read PCM sample from wave file
int32 IORiffWaveFileCached::readSample(int32* value, uint32 num, uint16 chan)
{
    if (isOutput) throw ErrorIntern("readSample: is output file");
    if (!cache)
    {
	// create cache:
	cache = new int32[cacheSamNum * getChannelNum()];
	if (!cache) throw ErrorSys(ENOMEM);

	// read cache:
	int32 rdsam = readCache(num);
	if (0 >= rdsam) return rdsam;
    }

    if (num >= cacheStart && num <= cacheEnd)
    {
	// sample is in cache:
	uint32 offset = (num - cacheStart) * getChannelNum() + chan;
	*value = cache[offset];
	return 1;
    }

    // read cache:
    int32 rdsam = readCache(num);
    if (0 >= rdsam) return rdsam;

    // get sample from cache:
    uint32 offset = (num - cacheStart) * getChannelNum() + chan;
    *value = cache[offset];

    return rdsam;
}


// write PCM sample into wave file
void IORiffWaveFileCached::writeSample(int32 value, uint32 num, uint16 chan)
{
    if (!isOutput) throw ErrorIntern("writeSample: is input file");

    if (cache && num >= cacheStart && num <= cacheEnd)
    {
	// sample fits in cache:
	uint32 offset = (num - cacheStart) * getChannelNum() + chan;
	value = cache[offset] = value;
	return;
    }

    // sample does not fit in cache:
    if (cache)
    {
	// write cache:
	writeCache();
    }
    else
    {
	// create cache:
	cache = new int32[cacheSamNum * getChannelNum()];
	if (!cache) throw ErrorSys(ENOMEM);
    }

    // try to read cache (just in case):
    int32 rdsam = readCache(num);
    cacheEnd = cacheStart + cacheSamNum - 1;
    if (0 >= rdsam) rdsam = 0;

    // clear unread part of cache:
    for (int32 curSam = rdsam; curSam < cacheSamNum; ++curSam)
    {
	for (uint16 curChan = 0; curChan < getChannelNum(); ++curChan)
	{
	    uint32 offset = curSam * getChannelNum() + curChan;
	    cache[offset] = 0;
	}
    }

    // put sample into cache:
    uint32 offset = (num - cacheStart) * getChannelNum() + chan;
    cache[offset] = value;
}



int32 IORiffWaveFileCached::readCache(uint32 startSamNum)
{
    // set read pointer to specified sample:
    setDataPos(startSamNum);

    // read cache:
    int32 rdsam = readSamples(cache, cacheSamNum);
    if (0 >= rdsam) return rdsam;
    if (rdsam > cacheSamNum) rdsam = cacheSamNum;

    cacheStart = startSamNum;
    cacheEnd = cacheStart + rdsam - 1;
    return rdsam;
}



void IORiffWaveFileCached::writeCache()
{
    // set write pointer to specified sample:
    setDataPos(cacheStart);

    // write cache:
    if (0 >= writeSamples(cache, cacheSamNum))
	throw ErrorBasic("Cannot write to " + std::string(waveName));
}



END_NAMESPACE
